import pyrql
from pyrql.query import And, Or, Filter
import uuid
from enum import Enum
from s3i import EventMessage
import time, copy


class EventFilterType(Enum):
    """
    Class for creating enumerations of event filter's type, which contains only RQL_FILTER used in S³I EventSystem
    """
    RQL_FILTER = 1


class EventManager:
    """
    Class for creating an event manager that manages an event system built in an edge based thing
    """

    def __init__(self, event_filter_type, ml40_model):
        """
        Constructor

        :param event_filter_type: type of event filter, by default (EventFilterType.RQL_FILTER)
        :param ml40_model: data model of monitored ml40 Thing
        :type ml40_model: dict
        """
        self.__sub_id_list = []
        self.__map_attr_sub = dict()
        self.__map_sub_attr = dict()
        # self.__fml40_model = benedict(fml40_model, keypath_separator=" ")
        self.__ml40_model = ml40_model
        self.__events = []
        self.__event_filter_type = event_filter_type

        if self.__event_filter_type == EventFilterType.RQL_FILTER:
            self.__rql = RQL(fml40=self.__ml40_model)

    def query_sub_id_list(self):
        """
        Function to query all subscription ids

        :return: subscription ids
        :rtype: list
        """
        return self.__sub_id_list

    def query_subscribed_event(self, sub_id):
        """
        Function to query a event via its subscription id

        :param sub_id: subscription id
        :type sub_id: str
        :return: event

        """
        for event in self.__events:
            if event.__dict__.get("sub_id") == sub_id:
                return event.__dict__

    @staticmethod
    def __generate_sub_id():
        """
        generate a S³I subscription id
        :return: subscription id
        :rtype: str

        """
        return "s3i:" + str(uuid.uuid4())

    def add_event(self, filter_expression: str, subscriber: str, subscriber_endpoint: str):
        """
        add event into event manager

        :param filter_expression: RQL expression
        :type filter_expression: str
        :param subscriber: id of event subscriber
        :type subscriber: str
        :param subscriber_endpoint: endpoint of event subscriber
        :type subscriber_endpoint: str

        :return: is event added or not, a generated event subscription id
        :rtype: bool, str

        """
        event = self.__add_event(filter_expression, subscriber, subscriber_endpoint)
        if event is None:
            return False, None
        for exist_event in self.__events:
            if str(exist_event.event_filter) == str(event.event_filter):
                if not set(event.subscribers) < set(exist_event.subscribers):
                    exist_event.subscribers += event.subscribers
                    exist_event.subscriber_endpoints += event.subscriber_endpoints
                return True, exist_event.sub_id
        self.__events.append(event)
        return True, event.sub_id

    def __add_event(self, filter_expression, subscriber, subscriber_endpoint):
        __attributes = []
        __event_filter = None
        if self.__event_filter_type == EventFilterType.RQL_FILTER:
            if not self.__rql.is_rql_valid(filter_expression):
                return None
            __event_filter = self.__rql.make_filter(rql_expr=filter_expression)
            __parsed_rql = self.__rql.parse(rql_expr=filter_expression)
            __attributes = self.__rql.get_attributes(__parsed_rql)
        __sub_id = self.__generate_sub_id()
        self.__sub_id_list.append(__sub_id)

        self.__map_sub_attr[__sub_id] = __attributes

        for __attr in __attributes:
            if self.__map_attr_sub.get(__attr) is None:
                self.__map_attr_sub[__attr] = list()
            self.__map_attr_sub[__attr].append(__sub_id)

        _event = Event(__sub_id, __attributes, __event_filter, [subscriber], [subscriber_endpoint],
                       ml40_model=self.__ml40_model)
        return _event

    def delete_event(self, subscription_id: str) -> bool:
        """
        delete event from event manager

        :param subscription_id: subscription id of event
        :type subscription_id: str
        :return: is event deleted
        :rtype: bool

        """
        for event in self.__events:
            if event.__dict__.get("sub_id") == subscription_id:
                self.__events.remove(event)
                return True
        return False

    def emit_event(self, publisher, sender_id):
        """
        emit events

        :param publisher: event publisher e.g. instantiated S³I Broker object that has a sending function
        :param sender_id: id of sender
        :type sender_id: str
        """
        while True:
            for event in self.__events:
                if event.is_filter_true:
                    event_message = event.generate_event_message(sender_id=sender_id)
                    publisher.send(receiver_endpoints=event.subscriber_endpoints, msg=event_message)

    def subject_updated(self, cur_ml40):
        """
        This function will be called if the subject's data model is changed

        :param path: has a changed value
        :type path: str
        """
        for event in self.__events:
            event(cur_ml40)
        """
        path = path.replace("/", " ")
        sub_ids = self.__map_attr_sub.get(path)
        if sub_ids is None:
            return None

        for sub_id in sub_ids:
            for event in self.__events:
                if event.__dict__.get("sub_id") == sub_id:
                    event(cur_ml40)
        """

class Event(object):
    """
    class for creating an event object
    """
    subscription_count = 0

    def __init__(self, sub_id, attributes, event_filter, subscribers, subscriber_endpoints,
                 ml40_model):
        """
        Constructor

        :param sub_id: subscription id
        :type sub_id: str
        :param attributes: in the event filter related attributes
        :type attributes: str
        :param event_filter: event filter, e.g. RQL filter
        :param subscribers: subscribers
        :type subscribers: list
        :param subscriber_endpoints: endpoints of subscribers
        :type subscriber_endpoints: list

        """
        self.sub_id = sub_id
        self.attributes = attributes
        self.event_filter = event_filter

        self.subscribers = subscribers
        self.subscriber_endpoints = subscriber_endpoints
        self.__class__.subscription_count += 1
        self._cur_dict = dict()
        self.__ml40_model = ml40_model
        self._is_filter_true = False

        self.__resGetValue = list()

    @property
    def is_filter_true(self):
        return self._is_filter_true

    def __del__(self):
        self.__class__.subscription_count -= 1

    def check_filter(self, cur_ml40):
        """
        Function to check if filter is set to true which means an event is triggered

        :param cur_ml40: current fml40 Json data model
        :type cur_ml40: dict
        :return: the status of filter
        :rtype: bool

        """

        self.__ml40_model = cur_ml40
        for attr in self.attributes:
            value = _uriToData(attr.replace(" ", "/"), self.__ml40_model)
            self._cur_dict.update({attr: value})
        if self.event_filter(self._cur_dict):
            self._is_filter_true = True

        else:
            self._is_filter_true = False

    def generate_event_message(self, sender_id):
        """
        Function to generate a S³I-B event message

        :param sender_id: id of sender
        :type sender_id: str
        :return: event message
        :rtype: dict

        """
        if not self._is_filter_true:
            return None

        event_message = EventMessage()
        content_dict = dict()
        for key in self._cur_dict.keys():
            content_dict[key.replace(" ", "/")] = self._cur_dict[key]

        event_message.fillEventMessage(sender=sender_id, receivers=self.subscribers,
                                       msg_id=str(uuid.uuid4()), sub_id=self.sub_id,
                                       timestamp=time.time(), content=content_dict)
        self._is_filter_true = False
        return event_message.msg.__str__()

    __call__ = check_filter


class RQL:
    """
    class for parsing, creating and checking rql filter
    """

    def __init__(self, fml40=dict):
        """
        Constructor

        :param fml40: fml40 data model
        :type fml40: dict


        """
        self._fml40 = fml40
        self._rql_logic_operator = ["and", "or"]
        self._rql_math_operator = ["eq", "ne", "gt", "ge", "lt", "ge", "in", "like", "exists"]

    @staticmethod
    def parse(rql_expr):
        """
        Function to parse rql expression

        :param rql_expr: rql expression
        :type rql_expr: str

        :return: parsed rql expression
        :rtype: str

        """
        try:
            rql_expr = rql_expr.replace("/", " ")
            parsed_rql = pyrql.parse(rql_expr)
            return parsed_rql
        except pyrql.RQLSyntaxError:
            return None

    def make_filter(self, rql_expr):
        """
        Function to create a RQL filter

        :param rql_expr: RQL expression
        :type rql_expr: str
        :return: RQL filter

        Returns:

        """
        parsed_rql = self.parse(rql_expr)
        if parsed_rql.get("name") in self._rql_math_operator:
            rql_filter = Filter(parsed_rql.get("name"), parsed_rql.get("args")[0], parsed_rql.get("args")[1])

        else:
            args = parsed_rql.get("args")
            rql_filter = self.__create_single_filter(args)
            for i in range(len(rql_filter) - 1):
                if parsed_rql.get("name") == "and":
                    rql_filter = And(rql_filter[i], rql_filter[i + 1])
                if parsed_rql.get("name") == "or":
                    rql_filter = Or(rql_filter[i], rql_filter[i + 1])
        return rql_filter

    def __create_single_filter(self, args, filters=list()):
        """
        Function to create a single filter which only contains one related variable in filter expression

        Args:
            args:
            filters:

        Returns:

        """
        for i in range(len(args)):
            if args[i].get("name") not in self._rql_logic_operator:
                filters.append(
                    Filter(args[i].get("name"), args[i].get("args")[0], args[i].get("args")[1]))
            else:
                filters.append(list())
                self._create_single_filter(args=args[i].get("args"), filters=filters[i])

                if args[i].get("name") == "and":

                    for j in range(len(filters[i]) - 1):
                        filters[i] = And(filters[i][j], filters[i][j + 1])

                if args[i].get("name") == "or":

                    for j in range(len(filters[i]) - 1):
                        filters[i] = Or(filters[i][j], filters[i][j + 1])

        return filters

    def get_attributes(self, parsed_rql):
        """
        Function to get all attributes listed in RQL expression

        :param parsed_rql: parsed rql expression
        :type parsed_rql: str
        :return: attributes
        :rtype: list

        """
        _attributes = list()
        if parsed_rql.get("name") in self._rql_math_operator:
            _attributes.append(parsed_rql.get("args")[0])

        else:
            args = parsed_rql.get("args")
            _attributes = self._get_attributes(args)
        return _attributes

    def _get_attributes(self, args):
        _attributes = list()
        for i in range(len(args)):
            if args[i].get("name") not in self._rql_logic_operator:
                _attributes.append(args[i].get("args")[0])
            else:
                _temp_attributes = self._get_attributes(args=args[i].get("args"))
                for temp in _temp_attributes:
                    _attributes.append(temp)
        return _attributes

    def is_rql_valid(self, rql_expr):
        """
        Function to check if a RQL expression is valid

        :param rql_expr: RQL expression
        :type rql_expr: str

        :return: status of rql validation
        :rtype: bool

        """
        parsed_rql = self.parse(rql_expr)
        if parsed_rql is None:
            return False

        attributes_list = self.get_attributes(parsed_rql)
        for attr in attributes_list:
            attr = attr.replace(" ", "/")
            try:
                var = _uriToData(uri=attr, ml40_model=self._fml40)
                if var == "Invalid attribute path":
                    return False
            except KeyError:
                return False
        return True

resGetValue = list()


def _uriToData(uri, ml40_model):
    """Returns a copy of the value found at uri.

    :param uri: Path to value
    :rtype: Feature

    """
    global resGetValue
    if uri == "":
        return ml40_model
    else:
        uri_list = uri.split("/")
        if uri_list[0] == "features":
            try:
                return ml40_model[uri]
            except KeyError:
                return "Invalid attribute path"

        try:
            _getValue(ml40_model, uri_list, ml40_model)
        except:
            return "Invalid attribute path"
        if resGetValue.__len__() == 0:
            return "Invalid attribute path"
        response = copy.deepcopy(resGetValue)
        resGetValue.clear()
        if response.__len__() == 1:
            return response[0]
        else:
            return response


def _getValue(source, uri_list, ml40_model):
    """Searches for the value specified by uri_list in source and stores
    the result in __resGetValue.

    :param source: Object that is scanned
    :param uri_list: List containing path

    """
    global resGetValue
    value = source[uri_list[0]]
    if uri_list.__len__() == 1:
        # if is ditto-feature
        if isinstance(value, str):
            try:
                stringValue_split = value.split(":")
                if stringValue_split[0] == "ditto-feature":
                    value = ml40_model["features"][stringValue_split[1]][
                        "properties"
                    ][uri_list[0]]
            except:
                pass
        resGetValue.append(value)
        return
    if isinstance(value, dict):
        # ??? uri_list.pop(0) better?!
        del uri_list[0]
        _getValue(value, uri_list, ml40_model)
    if isinstance(value, list):
        if isinstance(value[0], (str, int, float, bool, list)):
            return value
        if isinstance(value[0], dict):
            for item in value:
                if item["class"] == "ml40::Thing":
                    for i in item["roles"]:
                        if _findValue(i, uri_list[1]):
                            uri_list_1 = copy.deepcopy(uri_list)
                            del uri_list_1[:2]
                            _getValue(item, uri_list_1, ml40_model)
                    _f = _findValue({"identifier": item.get("identifier")}, uri_list[1]) or \
                         _findValue({"name": item.get("name")}, uri_list[1])
                    if _f:
                        uri_list_1 = copy.deepcopy(uri_list)
                        del uri_list_1[:2]
                        _getValue(item, uri_list_1, ml40_model)
                else:
                    if _findValue(item, uri_list[1]):
                        uri_list_1 = copy.deepcopy(uri_list)
                        del uri_list_1[:2]
                        if not uri_list_1:
                            resGetValue.append(item)
                            return
                        else:
                            _getValue(item, uri_list_1, ml40_model)
    if isinstance(value, (str, int, float, bool)):
        # if is ditto-feature
        if isinstance(value, str):
            try:
                stringValue_split = value.split(":")
                if stringValue_split[0] == "ditto-feature":
                    value = ml40_model["features"][stringValue_split[1]][
                        "properties"
                    ][uri_list[0]]
            except:
                pass
        resGetValue.append(value)


def _findValue(json, value):
    """Returns true if value has been found in json, otherwise returns false.

    :param json: dictionary
    :param value:
    :returns:
    :rtype:

    """

    # TODO: Simplify: value in json.values()
    for item in json:
        if json[item] == value:
            # print("Parameter: ", json[item])
            return True
    return False



import pgpy
import json
import gnupg
import time
from .key_pgp import Key
from abc import ABC, abstractmethod
from s3i.exception import raise_error_from_s3ib_msg, S3IBMessageError


class Message(ABC):
    """This is the abstract base class of an S3I message. It provides all functions concerning the signature and the encryption of an S3I message.
    """

    def __init__(self, msg):
        """Constructor of the abstract class.

        :param msg: PGP message, if empty or false type, a new PGP message is created.
        :type msg: str, bytes, pgpy.pgp.PGPMessage
        """
        if isinstance(msg, str or bytes):
            self._pgpMsg = pgpy.PGPMessage.from_blob(msg)
        elif isinstance(msg, pgpy.pgp.PGPMessage):
            self._pgpMsg = msg
        else:
            self._pgpMsg = pgpy.PGPMessage.new(str(msg))

    @property
    def pgpMsg(self):
        """PGP message
        """
        return self._pgpMsg

    @pgpMsg.setter
    def pgpMsg(self, value):
        self._pgpMsg = value

    def sign(self, secKey):
        """This function signs the message with the given private key. The signing is done as described in RFC 4880.

        :param secKey: Private PGP Key with which the signature is done. This key should be the secret key of the sender. It should be the counter part of the public key which is deposited in the S3I Directory or some other known database.
        :type secKey: pgpy.PGPKey
        :return: The signed PGP message
        :rtype: pgpy.PGPMessage 
        """
        self.pgpMsg |= secKey.sign(self.pgpMsg)
        return self.pgpMsg

    def verify(self, pubKey):
        """This function verifies the signature of the mesage with the given public key. If the signature is valid for the given key the function returns True.

        :param pubKey: Public PGP Key with which the signature is verified. This key should be the public key of the sender.
        :type pubKey: pgpy.PGPKey
        :return: Can be compared directly as a boolean to determine whether or not the specified signature verified.
        :rtype: pgpy.types.SignatureVerification
        """
        verif = pubKey.verify(self.pgpMsg)
        return verif

    def encrypt(self, pubKey):
        """This function encrypts the message with the given public key. The encryption is done as described in RFC 4880.

        :param pubKey: List of public PGP Keys with which the enryption is done. These keys should be the public keys of the receivers, which can be found in the S3I Directory or some other shared database.
        :type pubKey: list of pgpy.PGPKey
        :return: The encrypted PGP message is returned.
        :rtype: pgpy.PGPMessage
        """

        def encrypt_mess(pubKey, message, cipher, sessionKey):
            enc_msg = pubKey.encrypt(message, cipher=cipher, sessionkey=sessionkey)
            return enc_msg

        cipher = pgpy.constants.SymmetricKeyAlgorithm.AES256
        sessionkey = cipher.gen_key()
        for i in pubKey:
            self.pgpMsg = encrypt_mess(i.key, self.pgpMsg, cipher, sessionkey)
        return self.pgpMsg

    def decrypt(self, secKey):
        """This function decrypts the message with the given private key. 

        :param secKey: The private key of the receiver.
        :type secKey: pgpy.PGPKey
        :return: The decrypted PGP message is returned.
        :rtype: pgpy.PGPMessage 
        """
        dec = secKey.decrypt(self.pgpMsg)
        self.pgpMsg = dec
        print(
            "[S3I]["
            + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
            + "]: Decryption successful"
        )
        return dec

    def decryptAndVerify(self, secKey, fp, dir):
        """This function decrypts the message with the given private key and verifies the signature with the given public key. This function needs a gpg executable. If no gpg executable can be found the message is only decrypted but not verified. To install gpg on your computer go to https://gnupg.org/download/ and choose the right binary release for your operating system.

        :param secKey: The private key of the receiver.
        :type secKey: pgpy.PGPKey
        :param fp: Path to the home directory of the demo where a gpg folder is created.
        :type fp: str
        :param dir: An S3I Directory instance to look up the public key of the sender to verify the signature.
        :type dir: s3i.directory.Directory
        :return: The decrypted PGP message is returned. If the signature is not valid, the function returns None.
        :rtype: pgpy.PGPMessage 
        """
        try:
            gpg = gnupg.GPG(gnupghome=fp)
            import_result_sec = gpg.import_keys(secKey.key.__bytes__())
            decrypted_data = gpg.decrypt(self.pgpMsg.__bytes__())

            sender = json.loads(decrypted_data.data.decode("utf-8").replace("'", '"'))[
                "sender"
            ]
            key_blob = dir.getPublicKey(sender)
            pgpKey_pub = Key(key_str=key_blob)
            import_result_pub = gpg.import_keys(pgpKey_pub.key.__bytes__())

            decrypted_data = gpg.decrypt(self.pgpMsg.__bytes__())

            if decrypted_data.valid:
                print(
                    "[S3I]["
                    + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
                    + "]: The signature is valid."
                )
                print(
                    "[S3I]["
                    + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
                    + "]: Decryption successful"
                )
                self.msg = json.loads(
                    decrypted_data.data.decode("utf-8").replace("'", '"')
                )
                return decrypted_data.data
            else:
                return None
        except:
            print(
                "[S3I]["
                + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
                + "]: No Gpg executable can be found. The signature can not be verified. But the message will be decrypted."
            )
            return self.decrypt(secKey=secKey.key)


class UserMessage(Message):
    # S3I-B user message template
    _msg = {
        "sender": "",
        "identifier": "",
        "receivers": [],
        "messageType": "userMessage",
        "replyToEndpoint": "",
        "attachments": [],
        "subject": "",
        "text": "",
    }

    def __init__(
            self,
            json_in=None,
            msg_blob=None,
            sender=None,
            identifier=None,
            receivers=None,
            subject=None,
            text=None,
            replyToEndpoint=None,
            attachments=None,
    ):
        """Constructor

        :param json: JSON-formatted data to initialize the user message
        :type json: dict, string
        :param msg_blob: PGP message to initialize the user message
        :type msg_blob: string, bytes
        :param sender: id of the sender
        :type sender: string
        :param identifier: id of the message
        :type identitifer: string
        :param receivers: id of the receiver
        :type receivers: list of strings
        :param replyToEndpoint: endpoint of the sender
        :type replyToEndpoint: string
        :param subject: subject of the message
        :type subject: string
        :param text: text body of the message
        :type text: string
        :param attachments: List of dicts. Every dict has two keys: "filename" with value type str and "data" with value type BASE64-encoded str
        :type attachments: list
        """

        super().__init__(msg=msg_blob)
        if isinstance(json_in, dict):
            for k, v in json_in.items():
                self.msg[k] = v
        elif isinstance(json_in, str):
            x = json.loads(json_in)  # TODO check if it is template-conform
            self.msg = x

        fields = {
            "sender": sender,
            "identifier": identifier,
            "receivers": receivers,
            "subject": subject,
            "text": text,
            "replyToEndpoint": replyToEndpoint,
            "attachments": attachments,
        }
        for (key, value) in fields.items():
            if value:
                self.msg[key] = value
        if any(fields.values()):
            self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    @property
    def msg(self):
        """S3I-B user message template
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = value
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def fillUserMessage(
            self,
            senderUUID,
            receiverUUIDs,
            sender_endpoint,
            subject,
            text,
            msgUUID,
            attachments=None,
            attachments_json=None,
    ):
        """This function fills the user message with the given inputs.

        :param senderUUID: id of the sender
        :type senderUUID: string
        :param receiverUUID: id of the receiver
        :type receiverUUID: string
        :param sender_endpoint: endpoint of the sender
        :type sender_endpoint: string
        :param subject: subject of the message
        :type subject: string
        :param text: text body of the message
        :type text: string
        :param attachments: List of dicts. Every dict has two keys: "filename" with value type string and "data" with value type BASE64-encoded str
        :type attachments: list
        :param msgUUID: ID of the message (generated by client)
        :type msgUUID: string
        """

        self.msg["sender"] = senderUUID
        self.msg["identifier"] = msgUUID
        self.msg["receivers"] = receiverUUIDs
        self.msg["subject"] = subject
        self.msg["replyToEndpoint"] = sender_endpoint
        self.msg["text"] = text

        """
        add the input attachments (dict) to user message
        """
        if attachments is not None:
            len_attachments = len(attachments)
            self.msg["attachments"] = [
                {"filename": "", "data": ""} for i in range(len_attachments)
            ]
            for i in range(len_attachments):
                self.msg["attachments"][i]["filename"] = attachments[i]["filename"]
                self.msg["attachments"][i]["data"] = attachments[i]["data"]
        if attachments_json is not None:
            self.msg["attachments"] = attachments_json

        self.msg["messageType"] = "userMessage"
        raise_error_from_s3ib_msg(self.msg, S3IBMessageError)
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def convertPgpToMsg(self):
        """This function converts the PGP message format in the JSON-based S3I message format of a UserMessage.
        """
        json_msg = json.loads(self.pgpMsg.message.replace("'", '"'))
        self.fillUserMessage(
            senderUUID=json_msg["sender"],
            receiverUUIDs=json_msg["receivers"],
            sender_endpoint=json_msg["replyToEndpoint"],
            subject=json_msg["subject"],
            text=json_msg["text"],
            msgUUID=json_msg["identifier"],
            attachments_json=json_msg["attachments"],
        )


class ServiceRequest(Message):
    def __init__(self, json=None, msg_blob=None):
        """Constructor

        :param json: JSON-formatted data to initialize the service request
        :type json: dict, str
        :param msg_blob: PGP message to initialize the service request
        :type msg_blob: str, bytes
        """
        # S3I-B service request template
        self._msg = {
            "sender": "",
            "identifier": "",
            "receivers": [],
            "messageType": "serviceRequest",
            "replyToEndpoint": "",
            "serviceType": "",
            "parameters": {},
        }

        super().__init__(msg=msg_blob)
        if isinstance(json, dict):  # TODO check if it is template-conform
            for k, v in json.items():
                self.msg[k] = v
        elif isinstance(json, str):
            x = json.loads(json)
            self.msg = x

    @property
    def msg(self):
        """S3I-B service request template
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = value
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def fillServiceRequest(
            self,
            senderUUID,
            receiverUUIDs,
            sender_endpoint,
            serviceType,
            parameters,
            msgUUID,
    ):
        """This function fills the service request with the given inputs.

        :param senderUUID: id of the sender
        :type senderUUID: str
        :param receiverUUIDs: id of the receivers
        :type receiverUUIDs: list
        :param sender_endpoint: endpoint of the sender
        :type sender_endpoint: str
        :param serviceType: serviceType of the message
        :type serviceType: str
        :param parameters: parameters of the service request
        :type parameters: dict
        :param msgUUID: ID of the message (client-generated)
        :type msgUUID: str
        """

        self.msg["sender"] = senderUUID
        self.msg["identifier"] = msgUUID
        self.msg["receivers"] = receiverUUIDs
        self.msg["messageType"] = "serviceRequest"
        self.msg["replyToEndpoint"] = sender_endpoint
        self.msg["serviceType"] = serviceType

        if parameters is None:
            self.msg["parameters"] = ""
            print("[S3I]: Service request without parameters")

        else:
            self.msg["parameters"] = parameters
        raise_error_from_s3ib_msg(self.msg, S3IBMessageError)

    def convertPgpToMsg(self):
        """This function converts the PGP message format in the JSON-based S3I message format of a service request.
        """
        json_msg = json.loads(self.pgpMsg.message.replace("'", '"'))
        self.fillServiceRequest(
            senderUUID=json_msg["sender"],
            receiverUUIDs=json_msg["receivers"],
            sender_endpoint=json_msg["replyToEndpoint"],
            serviceType=json_msg["serviceType"],
            parameters=json_msg["parameters"],
            msgUUID=json_msg["identifier"],
        )


class ServiceReply(Message):
    def __init__(self, json=None, msg_blob=None):
        """Constructor

        :param json: JSON-formatted data to initialize the service reply
        :type json: dict, str
        :param msg_blob: PGP message to initialize the service reply
        :type msg_blob: str, bytes
        """
        # S3I-B service reply template
        self._msg = {
            "sender": "",
            "identifier": "",
            "receivers": [],
            "messageType": "serviceReply",
            "serviceType": "",
            "replyingToMessage": "",
            "results": {},
        }

        super().__init__(msg=msg_blob)
        if isinstance(json, dict):  # TODO check if it is template-conform
            for k, v in json.items():
                self.msg[k] = v
        elif isinstance(json, str):
            x = json.loads(json)
            self.msg = x

    @property
    def msg(self):
        """S3I-B service reply template
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = value
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def fillServiceReply(
            self, senderUUID, receiverUUIDs, serviceType, results, msgUUID, replyingToUUID
    ):
        """This function fills the service reply with the given inputs.

        :param senderUUID: id of the sender
        :type senderUUID: str
        :param receiverUUIDs: id of the receivers
        :type receiverUUIDs: list
        :param serviceType: serviceType of the message
        :type serviceType: str
        :param results: service results
        :type results: dict
        :param msgUUID: id of the message (client-generated)
        :type msgUUID: str
        :param replyingToUUID: id of the message this reply relates to
        :type replyingToUUID: str
        """

        self.msg["sender"] = senderUUID
        self.msg["identifier"] = msgUUID
        self.msg["receivers"] = receiverUUIDs
        self.msg["serviceType"] = serviceType
        self.msg["results"] = results
        self.msg["replyingToMessage"] = replyingToUUID
        raise_error_from_s3ib_msg(self.msg, S3IBMessageError)

    def convertPgpToMsg(self):
        """This function converts the PGP message format in the JSON-based S3I message format of a service reply.
        """
        json_msg = json.loads(self.pgpMsg.message.replace("'", '"'))
        self.fillServiceReply(
            senderUUID=json_msg["sender"],
            receiverUUIDs=json_msg["receivers"],
            replyingToUUID=json_msg["replyToEndpoint"],
            serviceType=json_msg["serviceType"],
            results=json_msg["results"],
            msgUUID=json_msg["identifier"],
        )


class GetValueRequest(Message):
    def __init__(self, json=None, msg_blob=None):
        """Constructor

        :param json: JSON-formatted data to initialize the get value request
        :type json: dict, str
        :param msg_blob: PGP message to initialize the get value request
        :type msg_blob: str, bytes
        """
        # S3I-B get value request template
        self._msg = {
            "sender": "",
            "identifier": "",
            "receivers": [""],
            "messageType": "getValueRequest",
            "replyToEndpoint": "",
            "attributePath": "",
        }
        super().__init__(msg=msg_blob)
        if isinstance(json, dict):  # TODO check if it is template-conform
            for k, v in json.items():
                self.msg[k] = v
        elif isinstance(json, str):
            x = json.loads(json)
            self.msg = x

    @property
    def msg(self):
        """S3I-B get value request template
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = value
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def fillGetValueRequest(
            self, senderUUID, receiverUUIDs: list, sender_endpoint, attributePath, msgUUID
    ):
        """This function fills the get value request with the given inputs.

        :param senderUUID: id of the sender
        :type senderUUID: str
        :param receiverUUIDs: id of the receivers
        :type receiverUUIDs: list
        :param sender_endpoint: endpoint of the sender
        :type sender_endpoint: str
        :param attributePath: path in the data model of the respective thing to get to the desired attribute
        :type attributePath: str
        :param msgUUID: ID of the message (client-generated)
        :type msgUUID: str
        """
        self.msg["sender"] = senderUUID
        self.msg["identifier"] = msgUUID
        self.msg["receivers"] = receiverUUIDs  # [0]["receiver"]
        self.msg["replyToEndpoint"] = sender_endpoint
        self.msg["attributePath"] = attributePath
        raise_error_from_s3ib_msg(self.msg, S3IBMessageError)

    def convertPgpToMsg(self):
        """This function converts the PGP message format in the JSON-based S3I message format of a get value request.
        """
        json_msg = json.loads(self.pgpMsg.message.replace("'", '"'))
        self.fillGetValueRequest(
            senderUUID=json_msg["sender"],
            receiverUUIDs=json_msg["receivers"],
            sender_endpoint=json_msg["replyToEndpoint"],
            attributePath=json_msg["attributePath"],
            msgUUID=json_msg["identifier"],
        )


class GetValueReply(Message):
    def __init__(self, json=None, msg_blob=None):
        """Constructor

        :param json: JSON-formatted data to initialize the get value reply
        :type json: dict, str
        :param msg_blob: PGP message to initialize the get value reply
        :type msg_blob: str, bytes
        """
        # S3I-B get value reply template
        self._msg = {
            "sender": "",
            "identifier": "",
            "receivers": [],
            "messageType": "getValueReply",
            "replyingToMessage": "",
            "value": {},
        }
        super().__init__(msg=msg_blob)
        if isinstance(json, dict):  # TODO check if it is template-conform
            for k, v in json.items():
                self.msg[k] = v
        elif isinstance(json, str):
            x = json.loads(json)
            self.msg = x

    @property
    def msg(self):
        """S3I-B get value reply template
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = value
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def fillGetValueReply(
            self, senderUUID, receiverUUIDs: list, results, msgUUID, replyingToUUID
    ):
        """This function fills the get value reply with the given inputs.

        :param senderUUID: id of the sender
        :type senderUUID: str
        :param receiverUUIDs: id of the receivers
        :type receiverUUIDs: list
        :param results: get value results
        :type results: dict
        :param msgUUID: id of the message (client-generated)
        :type msgUUID: str
        :param replyingToUUID: id of the message this reply relates to
        :type replyingToUUID: str
        """
        self.msg["sender"] = senderUUID
        self.msg["identifier"] = msgUUID
        self.msg["receivers"] = receiverUUIDs
        self.msg["replyingToMessage"] = replyingToUUID
        self.msg["value"] = results
        raise_error_from_s3ib_msg(self.msg, S3IBMessageError)

    def convertPgpToMsg(self):
        """This function converts the PGP message format in the JSON-based S3I message format of a get value reply.
        """
        json_msg = json.loads(self.pgpMsg.message.replace("'", '"'))
        self.fillGetValueReply(senderUUID=json_msg["sender"], receiverUUIDs=json_msg["receivers"],
                               replyingToUUID=json_msg["replyingToMessage"],
                               results=json_msg["value"], msgUUID=json_msg["identifier"])


class SetValueRequest(Message):
    def __init__(self, json=None, msg_blob=None):
        """Constructor

        :param json: JSON-formatted data to initialize the set value request
        :type json: dict, str
        :param msg_blob: PGP message to initialize the set value request
        :type msg_blob: str, bytes
        """
        # S3I-B set value request template
        self._msg = {
            "sender": "",
            "identifier": "",
            "receivers": [],
            "messageType": "setValueRequest",
            "replyToEndpoint": "",
            "attributePath": "",
            "newValue": 0
        }

        super().__init__(msg=msg_blob)
        if isinstance(json, dict):  # TODO check if it is template-conform
            for k, v in json.items():
                self.msg[k] = v
        elif isinstance(json, str):
            x = json.loads(json)
            self.msg = x

    @property
    def msg(self):
        """S3I-B set value request template
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = value
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def fillSetValueRequest(
            self,
            senderUUID,
            receiverUUIDs,
            sender_endpoint,
            attributePath,
            newValue,
            msgUUID,
    ):
        """This function fills the set value request with the given inputs.

        :param senderUUID: id of the sender
        :type senderUUID: str
        :param receiverUUIDs: id of the receivers
        :type receiverUUIDs: list
        :param sender_endpoint: endpoint of the sender
        :type sender_endpoint: str
        :param attributePath: attribute path
        :type attributePath: str
        :param newValue: new value
        :type newValue: any
        :param msgUUID: ID of the message (client-generated)
        :type msgUUID: str
        """

        self.msg["sender"] = senderUUID
        self.msg["identifier"] = msgUUID
        self.msg["receivers"] = receiverUUIDs
        self.msg["messageType"] = "setValueRequest"
        self.msg["replyToEndpoint"] = sender_endpoint
        self.msg["attributePath"] = attributePath
        self.msg["newValue"] = newValue

        raise_error_from_s3ib_msg(self.msg, S3IBMessageError)

    def convertPgpToMsg(self):
        """This function converts the PGP message format in the JSON-based S3I message format of a set value request.
        """
        json_msg = json.loads(self.pgpMsg.message.replace("'", '"'))
        self.fillServiceRequest(
            senderUUID=json_msg["sender"],
            receiverUUIDs=json_msg["receivers"],
            sender_endpoint=json_msg["replyToEndpoint"],
            attributePath=json_msg["attributePath"],
            newValue=json_msg["newValue"],
            msgUUID=json_msg["identifier"],
        )


class SetValueReply(Message):
    def __init__(self, json=None, msg_blob=None):
        """Constructor

        :param json: JSON-formatted data to initialize the set value reply
        :type json: dict, str
        :param msg_blob: PGP message to initialize the set value reply
        :type msg_blob: str, bytes
        """
        # S3I-B set value reply template
        self._msg = {
            "sender": "",
            "identifier": "",
            "receivers": [],
            "messageType": "setValueReply",
            "replyingToMessage": "",
            "ok": False,
        }

        super().__init__(msg=msg_blob)
        if isinstance(json, dict):  # TODO check if it is template-conform
            for k, v in json.items():
                self.msg[k] = v
        elif isinstance(json, str):
            x = json.loads(json)
            self.msg = x

    @property
    def msg(self):
        """S3I-B set value reply template
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = value
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def fillSetValueReply(
            self,
            senderUUID,
            receiverUUIDs,
            ok,
            replyingToUUID,
            msgUUID,
    ):
        """This function fills the set value reply with the given inputs.

        :param senderUUID: id of the sender
        :type senderUUID: str
        :param receiverUUIDs: id of the receivers
        :type receiverUUIDs: list
        :param replyingToUUID: reply to message
        :type replyingToUUID: str
        :param ok: indicates whether the value is set
        :type ok: str
        :param msgUUID: ID of the message (client-generated)
        :type msgUUID: str
        """

        self.msg["sender"] = senderUUID
        self.msg["identifier"] = msgUUID
        self.msg["receivers"] = receiverUUIDs
        self.msg["messageType"] = "setValueReply"
        self.msg["replyingToMessage"] = replyingToUUID
        self.msg["ok"] = ok

        raise_error_from_s3ib_msg(self.msg, S3IBMessageError)

    def convertPgpToMsg(self):
        """This function converts the PGP message format in the JSON-based S3I message format of a set value reply.
        """
        json_msg = json.loads(self.pgpMsg.message.replace("'", '"'))
        self.fillServiceRequest(
            senderUUID=json_msg["sender"],
            receiverUUIDs=json_msg["receivers"],
            sender_endpoint=json_msg["replyToEndpoint"],
            ok=json_msg["ok"],
            msgUUID=json_msg["identifier"],
        )


class EventSubscriptionRequest(Message):

    def __init__(self, json=None, msg_blob=None):
        """Constructor

        :param json: JSON-formatted data to initialize the event subscription request
        :type json: dict, str
        :param msg_blob: PGP message to initialize the event subscription request
        :type msg_blob: str, bytes
        """
        # S3I-B event subscription request template
        self._msg = {
            "sender": "",
            "identifier": "",
            "receivers": [],
            "messageType": "eventSubscriptionRequest",
            "replyToEndpoint": "",
            "RQL": ""
        }
        super().__init__(msg=msg_blob)
        if isinstance(json, dict):  # TODO check if it is template-conform
            for k, v in json.items():
                self.msg[k] = v
        elif isinstance(json, str):
            x = json.loads(json)
            self.msg = x

    @property
    def msg(self):
        """S3I-B event subscription request template
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = value
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def fillEventSubscriptionRequest(self, sender, receivers: list, msg_id, sender_endpoint, rql):
        """This function fills the event subscription request with the given inputs.

        :param sender: id of the sender
        :type sender: str
        :param receivers: id of the receivers
        :type receivers: list
        :param msg_id: id of the message (client-generated)
        :type msg_id: str
        :param sender_endpoint: endpoint of server that the message has been sent from
        :type sender_endpoint: str
        """
        self.msg["sender"] = sender
        self.msg["identifier"] = msg_id
        self.msg["receivers"] = receivers
        self.msg["replyToEndpoint"] = sender_endpoint
        self.msg["RQL"] = rql

    def convertPgpToMsg(self):
        """This function converts the PGP message format in the JSON-based S3I message format of a event subscription request.
        """
        json_msg = json.loads(self.pgpMsg.message.replace("'", '"'))
        self.fillEventSubscriptionRequest(sender=json_msg["sender"], receivers=json_msg["receivers"],
                                          msg_id=json_msg["identifier"], sender_endpoint=json_msg["replyToEndpoint"],
                                          rql=json_msg["RQL"])


class EventSubscriptionReply(Message):

    def __init__(self, json=None, msg_blob=None):
        """Constructor

        :param json: JSON-formatted data to initialize the event subscription reply
        :type json: dict, str
        :param msg_blob: PGP message to initialize the event subscription reply
        :type msg_blob: str, bytes
        """
        # S3I-B event subscription request template
        self._msg = {
            "sender": "",
            "identifier": "",
            "receivers": [],
            "messageType": "eventSubscriptionReply",
            "replyingToMessage": "",
            "subscription-id": "",
            "ok": bool
        }
        super().__init__(msg=msg_blob)
        if isinstance(json, dict):  # TODO check if it is template-conform
            for k, v in json.items():
                self.msg[k] = v
        elif isinstance(json, str):
            x = json.loads(json)
            self.msg = x

    @property
    def msg(self):
        """S3I-B event subscription reply template
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = value
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def fillEventSubscriptionReply(self, sender, receivers: list, msg_id, replying_to_message, subscription_id,
                                   status):
        """This function fills the event subscription request with the given inputs.

        :param sender: id of the sender
        :type sender: str
        :param receivers: id of the receivers
        :type receivers: list
        :param msg_id: id of the message (client-generated)
        :type msg_id: str
        :param replying_to_message: id of the message this reply relates to
        :type replying_to_message: str
        :param subscription_id: sub id
        :type subscription_id: str
        :param status: status of subscription
        :type status: bool
        """
        self.msg["sender"] = sender
        self.msg["identifier"] = msg_id
        self.msg["receivers"] = receivers
        self.msg["replyingToMessage"] = replying_to_message
        self.msg["subscription-id"] = subscription_id
        self.msg["ok"] = status

    def convertPgpToMsg(self):
        """This function converts the PGP message format in the JSON-based S3I message format of a event subscription reply.
        """
        json_msg = json.loads(self.pgpMsg.message.replace("'", '"'))
        self.fillEventSubscriptionRequest(sender=json_msg["sender"], receivers=json_msg["receivers"],
                                          msg_id=json_msg["identifier"],
                                          replying_to_message=json_msg["replyingToMessage"],
                                          subscription_id=json_msg["subscription-id"], status=json_msg["ok"])


class EventUnsubscriptionRequest(Message):

    def __init__(self, json=None, msg_blob=None):
        """Constructor

        :param json: JSON-formatted data to initialize the event unsubscribe request
        :type json: dict, str
        :param msg_blob: PGP message to initialize the event unsubscribe request
        :type msg_blob: str, bytes
        """
        # S3I-B event unsubscribe request template
        self._msg = {
            "sender": "",
            "identifier": "",
            "receivers": [],
            "messageType": "eventUnsubscriptionRequest",
            "replyToEndpoint": "",
            "subscription-id": ""
        }
        super().__init__(msg=msg_blob)
        if isinstance(json, dict):  # TODO check if it is template-conform
            for k, v in json.items():
                self.msg[k] = v
        elif isinstance(json, str):
            x = json.loads(json)
            self.msg = x

    @property
    def msg(self):
        """S3I-B event unsubscribe request template
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = value
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def fillEventUnsubscriptionRequest(self, sender, receivers: list, msg_id, sender_endpoint, sub_id):
        """This function fills the event unsubscribe request with the given inputs.

        :param sender: id of the sender
        :type sender: str
        :param receivers: id of the receivers
        :type receivers: list
        :param msg_id: id of the message (client-generated)
        :type msg_id: str
        :param sender_endpoint: id of the message this reply relates to
        :type sender_endpoint: str
        :param sub_id: event subscription id
        :type sub_id: str
        """
        self.msg["sender"] = sender
        self.msg["identifier"] = msg_id
        self.msg["receivers"] = receivers
        self.msg["replyToEndpoint"] = sender_endpoint
        self.msg["subscription-id"] = sub_id

    def convertPgpToMsg(self):
        """This function converts the PGP message format in the JSON-based S3I message format of a event unsubscribe reply.
        """
        json_msg = json.loads(self.pgpMsg.message.replace("'", '"'))
        self.fillEventSubscriptionRequest(sender=json_msg["sender"], receivers=json_msg["receivers"],
                                          msg_id=json_msg["identifier"], sender_endpoint=json_msg["replyToEndpoint"],
                                          sub_id=json_msg["subscription-id"])


class EventUnsubscriptionReply(Message):

    def __init__(self, json=None, msg_blob=None):
        """Constructor

        :param json: JSON-formatted data to initialize the event unsubscribe reply
        :type json: dict, str
        :param msg_blob: PGP message to initialize the event unsubscribe reply
        :type msg_blob: str, bytes
        """
        # S3I-B event unsubscribe reply template
        self._msg = {
            "sender": "",
            "identifier": "",
            "receivers": [],
            "messageType": "eventUnsubscriptionReply",
            "replyingToMessage": "",
            "subscription-id": "",
            "ok": bool
        }
        super().__init__(msg=msg_blob)
        if isinstance(json, dict):  # TODO check if it is template-conform
            for k, v in json.items():
                self.msg[k] = v
        elif isinstance(json, str):
            x = json.loads(json)
            self.msg = x

    @property
    def msg(self):
        """S3I-B event unsubscribe reply template
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = value
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def fillEventUnsubscriptionReply(self, sender, receivers: list, msg_id, replying_to_message, sub_id, status=bool):
        """This function fills the event unsubscription reply with the given inputs.

        :param sender: id of the sender
        :type sender: str
        :param receivers: id of the receivers
        :type receivers: list
        :param msg_id: id of the message (client-generated)
        :type msg_id: str
        :param replying_to_message: id of the message this reply relates to
        :type replying_to_message: str
        :param sub_id: event subscription id
        :type sub_id: str
        :param status: status of unsubscribe a event
        :type status: bool
        """
        self.msg["sender"] = sender
        self.msg["identifier"] = msg_id
        self.msg["receivers"] = receivers
        self.msg["replyToEndpoint"] = replying_to_message
        self.msg["subscription-id"] = sub_id
        self.msg["ok"] = status

    def convertPgpToMsg(self):
        """This function converts the PGP message format in the JSON-based S3I message format of a event unsubscribe reply.
        """
        json_msg = json.loads(self.pgpMsg.message.replace("'", '"'))
        self.fillEventSubscriptionRequest(sender=json_msg["sender"], receivers=json_msg["receivers"],
                                          msg_id=json_msg["identifier"],
                                          replying_to_message=json_msg["replyToEndpoint"],
                                          sub_id=json_msg["subscription-id"], status=json_msg["ok"])


class EventMessage(Message):
    def __init__(self, json=None, msg_blob=None):
        """Constructor

        :param json: JSON-formatted data to initialize the event message
        :type json: dict, str
        :param msg_blob: PGP message to initialize the event message
        :type msg_blob: str, bytes
        """
        # S3I-B event unsubscribe request template
        self._msg = {
            "sender": "",
            "identifier": "",
            "receivers": [],
            "messageType": "eventMessage",
            "subscription-id": "",
            "position": {
                "longtitude": "",
                "latitude": ""
            },
            "time": "",
            "content": ""
        }
        super().__init__(msg=msg_blob)
        if isinstance(json, dict):  # TODO check if it is template-conform
            for k, v in json.items():
                self.msg[k] = v
        elif isinstance(json, str):
            x = json.loads(json)
            self.msg = x

    @property
    def msg(self):
        """S3I-B event message template
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = value
        self.pgpMsg = pgpy.PGPMessage.new(str(self.msg))

    def fillEventMessage(self, sender, receivers: list, msg_id, sub_id, timestamp, content):
        """This function fills the event messages with the given inputs.

        :param sender: id of the sender
        :type sender: str
        :param receivers: id of the receivers
        :type receivers: list
        :param msg_id: id of the message (client-generated)
        :type msg_id: str
        :param sub_id: event subscription id
        :type sub_id: str
        :param timestamp: timestamp of current event
        :type timestamp: str
        :param content: content of event message
        :type content: str
        """
        self.msg["sender"] = sender
        self.msg["identifier"] = msg_id
        self.msg["receivers"] = receivers
        self.msg["subscription-id"] = sub_id
        self.msg["time"] = timestamp
        self.msg["content"] = content

    def convertPgpToMsg(self):
        """This function converts the PGP message format in the JSON-based S3I message format of a event message
        """
        json_msg = json.loads(self.pgpMsg.message.replace("'", '"'))
        self.fillEventMessage(sender=json_msg["sender"], receivers=json_msg["receivers"],
                              msg_id=json_msg["identifier"], sub_id=json_msg["subscription-id"],
                              timestamp=json_msg["time"], content=json_msg["content"])



import sys
import os
import functools
import ssl
import logging
import pika
import pika.exceptions
import abc
import requests
import urllib
import time
import asyncio
from pika.credentials import ExternalCredentials
from s3i.exception import raise_error_from_response, raise_error_from_s3ib_amqp, S3IBrokerRESTError, S3IBrokerAMQPError
from pika.adapters.asyncio_connection import AsyncioConnection

LOG_FORMAT = ('%(levelname) -10s %(asctime)s %(name) -30s %(funcName) '
              '-35s %(lineno) -5d: %(message)s')
LOGGER = logging.getLogger(__name__)


class BrokerMetaClass(metaclass=abc.ABCMeta):
    """
    Meta class for S³I Broker which is not allowed to be instantiated
    """

    @abc.abstractmethod
    def send(self, receiver_endpoints, msg, encrypted=False):
        """
        Sends an S3I-B message

        :param receiver_endpoints: endpoints of the receivers
        :type receiver_endpoints: list of str
        :param msg: message to be sent
        :type msg: str
        :param encrypted: if true, message will be sent encrypted, otherwise not.
        :type encrypted: bool

        """

    @abc.abstractmethod
    def receive_once(self, queue):
        """
        Receive one S3I-B message and do not wait for more messages.

        :param queue: queue which starts a listener in order to receive a single message
        :type queue: str
        """


class BrokerREST(BrokerMetaClass):
    """
    Class Broker REST contains functions to connect to S3I Broker via HTTP REST API, and send and receive messages

    """

    def __init__(self, token, url="https://broker.s3i.vswf.dev/"):
        """
        Constructor

        :param token: Access Token issued from S³I IdentityProvider
        :type token: str
        :param url: url of S³I Broker API
        :type url: str

        """
        self._token = token
        self._url = url
        self.headers = {'Content-Type': 'application/json',
                        'Authorization': 'Bearer ' + self.token}
        self.headers_encrypted = {'Content-Type': 'application/pgp-encrypted',
                                  'Authorization': 'Bearer ' + self.token}

    @property
    def token(self):
        """Returns the JWT currently in use.

        :returns: JWT-Token
        :rtype: str

        """
        return self._token

    @token.setter
    def token(self, new_token):
        """Sets the value of the object's token property to new_token.

        :param new_token: JWT
        :type new_token: str
        """
        self._token = new_token

    def send(self, receiver_endpoints, msg, encrypted=False):
        """
        Send a S³I-B message via S³I Broker API
        :param receiver_endpoints: endpoints of the receivers
        :type receiver_endpoints: list of str
        :param msg: message to be sent
        :type msg: str
        :param encrypted: if true, message will be sent encrypted, otherwise not.
        :type encrypted: bool

        """
        end_points_encoded = []
        for receiver_endpoint in receiver_endpoints:
            encoded = urllib.parse.quote(receiver_endpoint, safe="")
            end_points_encoded.append(encoded)

        endpoints = ",".join(end_points_encoded)
        url = "{}{}".format(self._url, endpoints)
        headers = self.headers
        if encrypted:
            headers = self.headers_encrypted
        response = requests.post(url=url, headers=headers, data=msg)
        raise_error_from_response(response, S3IBrokerRESTError, 201)

    def receive_once(self, queue):
        """
        Receive one S3I-B message and do not wait for more messages.

        :param queue: queue which starts a listener in order to receive a single message
        :type queue: str
        :return: received S3I-B message
        """

        queue_encoded = urllib.parse.quote(queue, safe="")
        url = "{}{}".format(self._url, queue_encoded)
        response = requests.get(url=url, headers=self.headers)
        response_json = raise_error_from_response(response, S3IBrokerRESTError, 200)
        return response_json


class Broker(BrokerMetaClass):
    """
    Class Broker contains functions to connect to S3I Broker via AMQP, and send and receive messages, using Pika library

    """
    CONTENT_TYPE = "application/json"
    HOST = "rabbitmq.s3i.vswf.dev"

    def __init__(self, auth_form, content_type=CONTENT_TYPE,
                 username=None, password=None, x509_path=None,
                 host=HOST):
        """
        Constructor
        :param auth_form: method to authentication against Broker. 'Username/Password' and 'X509' are valid
        :param content_type: accepted content type of messages. By default 'application/json'
        :param username: Username used for plain authentication. For OAuth2 can be empty
        :param password: Password used for plain authentication. For OAuth2 must be access token
        :param x509_path: Local path for X509 certificate, if 'X509' is chosen as 'auth_form'
        :param host: Host url of Broker
        """

        self.__auth_form = auth_form
        self.__username = username
        self.__password = password
        self.__content_type = content_type
        self.__x509_path = x509_path
        self.__host = host

        self._queue = None
        self._callback = None

        self.receiver = None
        self.receiver_once = None
        self.publisher = None

        self._reconnect_delay = 0

    def receive(self, queue, callback=None):
        """
        Starts consuming to receive messages with a callback function. You need to acknowledge the message in callback.

        :param queue: endpoint (queue) of receiver defined in S3I Broker,
        :type queue: str
        :param callback: callback function
        """
        self._queue = queue
        self._callback = callback

        new_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(new_loop)

        self.receiver = AsyncReceiver(
            connection_parameter=self.build_conn_par(),
            queue=self._queue, callback=self._callback)
        while True:
            try:
                self.receiver.run()
            except KeyboardInterrupt:
                self.receiver.stop()
                break
            self.maybe_reconnect()

    def maybe_reconnect(self, token=None):
        """
        Try to reconnect to the Broker, if the connection is aborted.
        :param token: new token
        :return: None
        """
        self.__password = token
        if self.receiver.should_reconnect:
            self.receiver.stop()
            reconnect_delay = self._get_reconnect_delay()
            LOGGER.info('Reconnecting after %d seconds', reconnect_delay)
            time.sleep(reconnect_delay)
            self.receiver = AsyncReceiver(
                connection_parameter=self.build_conn_par(),
                queue=self._queue, callback=self._callback)

    def _get_reconnect_delay(self):
        """
        Get a delay for reconnection, by default 0 sec.
        :return: reconnect delay
        """
        if self.receiver.was_consuming:
            self._reconnect_delay = 0
        else:
            self._reconnect_delay += 1
        if self._reconnect_delay > 30:
            self._reconnect_delay = 30
        return self._reconnect_delay

    def receive_once(self, queue):
        """
        Receives message once via AMQP

        :param queue: endpoint (queue) of receiver defined in S3I Broker
        :type queue: str
        :return: message body
        :rtype: bytearray
        """
        self.receiver_once = BrokerOnceReceiver(connection_parameter=self.build_conn_par())
        return self.receiver_once.receive_once(queue=queue)

    def send(self, receiver_endpoints, msg, encrypted=False):
        """
        Sends a message via AMQP

        :param receiver_endpoints: endpoints (queues) of receivers
        :type receiver_endpoints: list
        :param msg: message
        :type msg: str
        :param encrypted: whether the message is sent encrypted. TODO
        :type encrypted: bool


        """
        if not isinstance(receiver_endpoints, list):
            raise S3IBrokerAMQPError(error_msg="receiver endpoints must be list")
        self.publisher = BrokerOncePublisher(connection_parameter=self.build_conn_par())
        self.publisher.send(receiver_endpoints, msg)

    def build_conn_par(self):
        """
        Authentication at S3I Broker has two different methods
        X.509: Authentication with X.509 certificate which includes the user information or \n
        Username/Password: the corresponding user must have been registered in S3I Broker. To authenticate with an access token, using Username/Password \n
        :return: connection parameters for the S3I Broker; used in function buildConn(self, connpara)
        :rtype: pika connectionParameters, refer to ConnectionParameters_

        """

        auth_list = ["X509", "Username/Password"]
        try:
            auth_list.index(self.__auth_form)
        except ValueError as e:
            sys.exit("[x] invalid authentication form, please check")
        else:
            if self.__auth_form == "Username/Password":
                credentials = pika.PlainCredentials(
                    username=self.__username, password=self.__password)
                conn_par = pika.ConnectionParameters(
                    host=self.__host, virtual_host='s3i', credentials=credentials)
                return conn_par
            elif self.__auth_form == "X509":
                ca_cer_path = os.path.join(
                    self.__x509_path, "ca_certificate.pem")
                context = ssl.create_default_context(cafile=ca_cer_path)
                client_cer_path = os.path.join(
                    self.__x509_path, "client_certificate.pem")
                client_key_path = os.path.join(
                    self.__x509_path, "client_key.pem")
                context.load_cert_chain(client_cer_path, client_key_path)
                ssl_options = pika.SSLOptions(context, self.__host)
                conn_par = pika.ConnectionParameters(port=5671, host=self.__host, ssl_options=ssl_options,
                                                     virtual_host='s3i', credentials=ExternalCredentials())
                return conn_par


class AsyncReceiver:
    """
    Class contains functions for a asynchronous Broker consumer
    """
    def __init__(self, connection_parameter, queue, callback):
        """
        Constructor

        :param connection_parameter: connection parameter, built via function 'build_conn_par' in Class 'Broker'
        :param queue: message queue
        :param callback: callback function for message receiving
        """
        self._connection_parameter = connection_parameter
        self._connection = None
        self._channel = None

        self._closing = False
        self._consuming = False
        self._consumer_tag = None

        self.should_reconnect = False
        self.was_consuming = False

        self._queue = queue
        self._callback = callback
        self._prefetch_count = 1

    @property
    def connection_parameter(self):
        return self._connection_parameter

    @connection_parameter.setter
    def connection_parameter(self, value):
        self._connection_parameter = value

    @property
    def connection(self):
        return self._connection

    @connection.setter
    def connection(self, value):
        self._connection = value

    @property
    def consuming(self):
        return self._consuming

    @consuming.setter
    def consuming(self, value):
        self._consuming = value

    @property
    def consumer_tag(self):
        return self._consumer_tag

    def connect(self):
        """This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika.
        :rtype: pika.adapters.asyncio_connection.AsyncioConnection
        """
        LOGGER.info('Connecting to S3I Broker Receiver')
        return AsyncioConnection(
            parameters=self.connection_parameter,
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed)

    def close_connection(self):
        """
        This method disconnects to RabbitMQ
        :return: None
        """
        self._consuming = False
        if self._connection.is_closing or self._connection.is_closed:
            LOGGER.info('Connection is closing or already closed')
        else:
            LOGGER.info('Closing connection')
            self._connection.close()

    def on_connection_open(self, _unused_connection):
        """This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.
        :param pika.adapters.asyncio_connection.AsyncioConnection _unused_connection:
           The connection
        """
        LOGGER.info('Connection opened')
        self.open_channel()

    def on_connection_open_error(self, _unused_connection, err):
        """This method is called by pika if the connection to RabbitMQ
        can't be established.
        :param pika.adapters.asyncio_connection.AsyncioConnection _unused_connection:
           The connection
        :param Exception err: The error
        """
        LOGGER.error('Connection open failed: %s', err)
        self.reconnect()

    def on_connection_closed(self, _unused_connection, reason):
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.
        :param pika.connection.Connection connection: The closed connection obj
        :param Exception reason: exception representing reason for loss of
            connection.
        """
        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
        else:
            LOGGER.warning('Connection closed, reconnect necessary: %s', reason)
            self.reconnect()

    def reconnect(self):
        """Will be invoked if the connection can't be opened or is
        closed. Indicates that a reconnect is necessary then stops the
        ioloop.
        """
        self.should_reconnect = True
        self.stop()

    def open_channel(self):
        """Open a new channel with RabbitMQ by issuing the Channel.Open RPC
        command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.
        """
        LOGGER.info('Creating a new channel')
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.
        Since the channel is now open, we'll declare the exchange to use.
        :param pika.channel.Channel channel: The channel object
        """
        LOGGER.info('Channel opened')
        self._channel = channel
        self.add_on_channel_close_callback()
        self.set_qos()

    def add_on_channel_close_callback(self):
        """This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel.
        """
        LOGGER.info('Adding channel close callback')
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reason):
        """Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.
        :param pika.channel.Channel: The closed channel
        :param Exception reason: why the channel was closed
        """
        LOGGER.warning('Channel %i was closed: %s', channel, reason)
        self.close_connection()

    def set_qos(self):
        """This method sets up the consumer prefetch to only be delivered
        one message at a time. The consumer must acknowledge this message
        before RabbitMQ will deliver another one. You should experiment
        with different prefetch values to achieve desired performance.
        """
        self._channel.basic_qos(
            prefetch_count=self._prefetch_count, callback=self.on_basic_qos_ok)

    def on_basic_qos_ok(self, _unused_frame):
        """Invoked by pika when the Basic.QoS method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.
        :param pika.frame.Method _unused_frame: The Basic.QosOk response frame
        """
        LOGGER.info('QOS set to: %d', self._prefetch_count)
        self.start_consuming()

    def start_consuming(self):
        """This method sets up the consumer by first calling
        add_on_cancel_callback so that the object is notified if RabbitMQ
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with RabbitMQ. We keep the value to use it when we want to
        cancel consuming. The on_message method is passed in as a callback pika
        will invoke when a message is fully received.
        """
        LOGGER.info('Issuing consumer related RPC commands')
        self.add_on_cancel_callback()
        if self._callback is None:
            self._callback = self.on_message
        self._consumer_tag = self._channel.basic_consume(
            self._queue, self._callback)
        self.was_consuming = True
        self._consuming = True

    def add_on_cancel_callback(self):
        """Add a callback that will be invoked if RabbitMQ cancels the consumer
        for some reason. If RabbitMQ does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.
        """
        LOGGER.info('Adding consumer cancellation callback')
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.
        :param pika.frame.Method method_frame: The Basic.Cancel frame
        """
        LOGGER.info('Consumer was cancelled remotely, shutting down: %r',
                    method_frame)
        if self._channel:
            self._channel.close()

    def on_message(self, _unused_channel, basic_deliver, properties, body):
        """Invoked by pika when a message is delivered from RabbitMQ. The
        channel is passed for your convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.
        :param pika.channel.Channel _unused_channel: The channel object
        :param pika.Spec.Basic.Deliver: basic_deliver method
        :param pika.Spec.BasicProperties: properties
        :param bytes body: The message body
        """
        LOGGER.info('Received message # %s from %s: %s',
                    basic_deliver.delivery_tag, properties.app_id, body)
        self.acknowledge_message(basic_deliver.delivery_tag)

    def acknowledge_message(self, delivery_tag):
        """Acknowledge the message delivery from RabbitMQ by sending a
        Basic.Ack RPC method for the delivery tag.
        :param int delivery_tag: The delivery tag from the Basic.Deliver frame
        """
        LOGGER.info('Acknowledging message %s', delivery_tag)
        self._channel.basic_ack(delivery_tag)

    def stop_consuming(self):
        """Tell RabbitMQ that you would like to stop consuming by sending the
        Basic.Cancel RPC command.
        """
        if self._channel:
            LOGGER.info('Sending a Basic.Cancel RPC command to RabbitMQ')
            cb = functools.partial(
                self.on_cancelok, userdata=self._consumer_tag)
            self._channel.basic_cancel(self._consumer_tag, cb)

    def on_cancelok(self, _unused_frame, userdata):
        """This method is invoked by pika when RabbitMQ acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.
        :param pika.frame.Method _unused_frame: The Basic.CancelOk frame
        :param str|unicode userdata: Extra user data (consumer tag)
        """
        self._consuming = False
        LOGGER.info(
            'RabbitMQ acknowledged the cancellation of the consumer: %s',
            userdata)
        self.close_channel()

    def close_channel(self):
        """Call to close the channel with RabbitMQ cleanly by issuing the
        Channel.Close RPC command.
        """
        LOGGER.info('Closing the channel')
        self._channel.close()

    def run(self):
        """Run the example consumer by connecting to RabbitMQ and then
        starting the IOLoop to block and allow the AsyncioConnection to operate.
        """
        self._connection = self.connect()
        self._connection.ioloop.run_forever()

    def stop(self):
        """Cleanly shutdown the connection to RabbitMQ by stopping the consumer
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when CTRL-C is pressed raising a KeyboardInterrupt exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with RabbitMQ. All of the commands issued prior to starting
        the IOLoop will be buffered but not processed.
        """
        if not self._closing:
            self._closing = True
            LOGGER.info('Stopping')
            if self._consuming:
                self.stop_consuming()
                try:
                    self._connection.ioloop.run_forever()
                except RuntimeError:
                    pass
            else:
                self._connection.ioloop.stop()
            LOGGER.info('Stopped')


class BrokerOncePublisher:
    """
    Class contains functions for a message publisher
    """
    def __init__(self, connection_parameter):
        """
        Constructor

        :param connection_parameter: connection parameter, built via 'build_conn_par' in the class 'Broker'
        """
        self._connection_parameter = connection_parameter
        self._connection = None
        self._channel = None

    @property
    def connection_parameter(self):
        return self._connection_parameter

    @connection_parameter.setter
    def connection_parameter(self, value):
        self._connection_parameter = value

    def connect(self):
        """
        This method connects to RabbitMQ via BlockingConnection

        :return: None
        """
        self._connection = raise_error_from_s3ib_amqp(pika.BlockingConnection, S3IBrokerAMQPError,
                                                      self._connection_parameter)
        self._channel = raise_error_from_s3ib_amqp(self._connection.channel, S3IBrokerAMQPError)
        self._channel.confirm_delivery()

    def send(self, receiver_endpoints, msg):
        """
        This method sends a message to one or more endpoints (queues). After the sending, the connection will be aborted

        :param receiver_endpoints: queues of receivers
        :type receiver_endpoints: list
        :param msg: message
        :return: None
        """
        self.connect()
        if self._channel is None or not self._channel.is_open:
            raise S3IBrokerAMQPError(error_msg="Broker Channel is not open")
        for endpoint in receiver_endpoints:
            raise_error_from_s3ib_amqp(self._channel.basic_publish, S3IBrokerAMQPError, exchange="demo.direct",
                                       routing_key=endpoint,
                                       properties=pika.BasicProperties(content_type="application/json",
                                                                       delivery_mode=2),
                                       body=msg, mandatory=True)
        self.stop_publish()

    def stop_publish(self):
        """
        This method disconnects to RabbitMQ

        :return: None
        """
        if self._channel:
            self._channel.close()
        if self._connection:
            if self._connection.is_closed:
                return
            self._connection.close()


class BrokerOnceReceiver:
    """
    Class contains functions for a broker receiver, which only receives one message before disconnecting
    """
    def __init__(self, connection_parameter):
        """
        Constructor

        :param connection_parameter: Connection parameter, built via 'built_conn_par' in the class 'Broker'
        """
        self._connection_parameter = connection_parameter
        self._channel = None
        self._connection = None

    @property
    def connection_parameter(self):
        return self._connection_parameter

    @connection_parameter.setter
    def connection_parameter(self, value):
        self._connection_parameter = value

    def connect(self):
        """
        This method connects to RabbitMQ via Blocking connection
        :return: None
        """
        self._connection = raise_error_from_s3ib_amqp(pika.BlockingConnection, S3IBrokerAMQPError,
                                                      self._connection_parameter)
        self._channel = raise_error_from_s3ib_amqp(self._connection.channel, S3IBrokerAMQPError)

    def receive_once(self, queue):
        """
        This method receive only one message before disconnecting

        :param queue: endpoint of the receiver
        :return: message body
        :rtype: bytearray
        """
        self.connect()
        method_frame, header_frame, body = raise_error_from_s3ib_amqp(self._channel.basic_get, S3IBrokerAMQPError,
                                                                      queue)
        if method_frame:
            raise_error_from_s3ib_amqp(self._channel.basic_ack, S3IBrokerAMQPError, method_frame.delivery_tag)
        self.stop_receive()
        return body

    def stop_receive(self):
        """
        This method disconnects to RabbitMQ

        :return: None
        """
        if self._channel:
            self._channel.close()
        if self._connection:
            if self._connection.is_closed:
                return
            self._connection.close()


